/* d_diode.cc
 *$Header: /al/acs/src/RCS/d_diode.cc,v 9.26 95/10/31 16:11:40 al Exp $
 * diode model.
 * netlist syntax:
 * device:  dxxxx n+ n- mname <area> <off> <ic=vd> <model-card-args>
 * model:   .model mname D <args>
 */
#include "ap.h"
#include "m_matrix.h"
#include "d_admit.h"
#include "d_cap.h"
#include "d_diode.h"
#include "e_aux.h"
#include "error.h"
#include "io.h"
#include "s__.h"
#include "u_opt.h"
#include "u_status.h"
#include "util.h"
/*--------------------------------------------------------------------------*/
//		MODEL_DIODE::MODEL_DIODE(const char*);
//	void	MODEL_DIODE::parse(CS& cmd);
// 	void	MODEL_DIODE::print(int,int)const;
//		DIODE_COMMON::DIODE_COMMON()
//		DEV_DIODE::DEV_DIODE();
//		DEV_DIODE::DEV_DIODE(const DEV_DIODE& p);
//	void	DEV_DIODE::parse(CS& cmd);
// 	void	DEV_DIODE::print(int,int)const;
//	void	DEV_DIODE::expand();
//	double	DEV_DIODE::probe_tr_num(const char *what)const;
//	void	DEV_DIODE::evaldiode(CARD*);
//	void	DEV_DIODE::diodecap(CARD*);
/*--------------------------------------------------------------------------*/
int DEV_DIODE::Count = 0;
/*--------------------------------------------------------------------------*/
MODEL_DIODE::MODEL_DIODE(const char *name):MODEL_BASE(name)
{
  js   = 1e-14;
  rs   = 0.0;
  n    = 1.0;
  tt   = 0.0;
  cjo  = 0.0;
  pb   = 1.0;
  mj   = 0.5;
  eg   = 1.11;
  xti  = 3.0;
  kf   = 0.0;
  af   = 1.0;
  fc   = 0.5;
  bv   = 0.0;	/* infinity */
  ibv  = 1e-3;
  cjsw = 0.0;
  mjsw = 0.33;
  fcpb = fc * pb;
}
/*--------------------------------------------------------------------------*/
void MODEL_DIODE::parse(CS& cmd)
{
  cmd.skiparg();			/* skip known ".model" */
  cmd.ctostr(label, LABELEN, TOKENTERM);
  cmd.skiparg();			/* skip known "d" */
  cmd.skiplparen();
  for (;;){
    if (cmd.argparse(REPEAT,
	"IS",	aUDOUBLE,	&js,
	"RS",	aUDOUBLE,	&rs,
	"N",	aUDOUBLE,	&n,
	"TT",	aUDOUBLE,	&tt,
	"CJo",	aUDOUBLE,	&cjo,
	"VJ",	aUDOUBLE,	&pb,
	"PB",	aUDOUBLE,	&pb,
	"Mj",	aUDOUBLE,	&mj,
	"EGap",	aUDOUBLE,	&eg,
	"XTI",	aUDOUBLE,	&xti,
	"KF",	aUDOUBLE,	&kf,
	"AF",	aUDOUBLE,	&af,
	"FC",	aUDOUBLE,	&fc,
	"BV",	aUDOUBLE,	&bv,
	"IBV",	aUDOUBLE,	&ibv,
	"CJSw",	aUDOUBLE,	&cjsw,
	"MJSw",	aUDOUBLE,	&mjsw,
	"")){
      ;
    }else{
      cmd.skiprparen();
      cmd.check(bWARNING);
      break;
    }
  }
  fcpb = fc * pb;
}
/*--------------------------------------------------------------------------*/
void MODEL_DIODE::print(int where, int detail)const
{
  mprintf(where, ".model  %s  d  (", label);
  mprintf(where, " is=%s ",  ftos(js,  "", 7, 0));
  mprintf(where, " rs=%s ",  ftos(rs,  "", 7, 0));
  mprintf(where, " n=%s ",   ftos(n,   "", 7, 0));
  mprintf(where, " tt=%s ",  ftos(tt,  "", 7, 0));
  mprintf(where, " cjo=%s ", ftos(cjo, "", 7, 0));
  mprintf(where, " vj=%s ",  ftos(pb,  "", 7, 0));
  mprintf(where, " m=%s ",   ftos(mj,  "", 7, 0));
  mprintf(where, " eg=%s ",  ftos(eg,  "", 7, 0));
  mprintf(where, " xti=%s ", ftos(xti, "", 7, 0));
  mprintf(where, " kf=%s ",  ftos(kf,  "", 7, 0));
  mprintf(where, " af=%s ",  ftos(af,  "", 7, 0));
  mprintf(where, " fc=%s ",  ftos(fc,  "", 7, 0));
  mprintf(where, " bv=%s ",  ftos(bv,  "", 7, 0));
  mprintf(where, " ibv=%s ", ftos(ibv, "", 7, 0));
  if (cjsw != 0.){
    mprintf(where, " cjsw=%s ", ftos(cjsw,"", 7, 0));
    mprintf(where, " mjsw=%s ", ftos(mjsw,"", 7, 0));
  }
  mprintf(where, ")\n");
}
/*--------------------------------------------------------------------------*/
DIODE_COMMON::DIODE_COMMON()
{
  strcpy(modelname, "-error-");
  is	= NOT_INPUT;
  rs	= NOT_INPUT;
  cj	= NOT_INPUT;
  cjsw	= NOT_INPUT;
  area	= 1.0;
  perim	= 0.0;
  ic	= NOT_INPUT;
  off	= FALSE;
  calc.is = calc.rs = calc.cj = calc.cjsw = FALSE;
}
/*--------------------------------------------------------------------------*/
DEV_DIODE::DEV_DIODE()
{
  common = new DIODE_COMMON;
  devclass = OTHERDEVICE;
  Yj = NULL;
  Cj = NULL;
  region = UNKNOWN;
  ++Count;
}
/*--------------------------------------------------------------------------*/
DEV_DIODE::DEV_DIODE(const DEV_DIODE& p):BASE_SUBCKT(p)
{
  Yj = NULL;
  Cj = NULL;
  region = UNKNOWN;
  ++Count;
}
/*--------------------------------------------------------------------------*/
void DEV_DIODE::parse(CS& cmd)
{
  DIODE_COMMON* c = (DIODE_COMMON*)common;
  assert(c);
  
  parselabel(cmd);
  parsenodes(cmd,NUMNODES);
  cmd.ctostr(c->modelname, LABELEN, TOKENTERM);
  c->model = NULL;
  if (cmd.is_pfloat()){
    c->area = cmd.ctopf();
  }
  for (;;){
    if (cmd.argparse(REPEAT,
	"Area",	aUDOUBLE,	&c->area,
	"Perim",aUDOUBLE,	&c->perim,
	"IC",	aDOUBLE,	&c->ic,
	"OFF",	aENUM,		&c->off,	TRUE,
	"IS",	aUDOUBLE,	&c->is,
	"Rs",	aUDOUBLE,	&c->rs,
	"Cjo",	aUDOUBLE,	&c->cj,
	"CJSW",	aUDOUBLE,	&c->cjsw,
	"")){
      ;
    }else{
      cmd.check(bWARNING);
      break;
    }
  }
}
/*--------------------------------------------------------------------------*/
void DEV_DIODE::print(int where, int detail)const
{
  DIODE_COMMON* c = (DIODE_COMMON*)common;
  assert(c);

  printlabel(where);
  printnodes(where,NUMNODES);
  mprintf(where, " %s ", c->modelname);

  mprintf(where,    " %s ", ftos(c->area,"", 7, 0));
  if (c->perim != 0.)
    mprintf(where,"perim=%s ",ftos(c->perim,"", 7, 0));
  if (c->off)
    mprintf(where, " off ");
  if (c->ic != NOT_INPUT)
    mprintf(where, " ic=%s ", ftos(c->ic,  "", 7, 0));
  if (!c->calc.is  &&  c->is != NOT_INPUT)
    mprintf(where, " is=%s ", ftos(c->is,  "", 7, 0));
  if (!c->calc.rs  &&  c->rs != NOT_INPUT)
    mprintf(where, " rs=%s ", ftos(c->rs,  "", 7, 0));
  if (!c->calc.cj  &&  c->cj != NOT_INPUT)
    mprintf(where, " cj=%s ", ftos(c->cj,  "", 7, 0));
  if (!c->calc.cjsw  &&  c->cjsw != NOT_INPUT)
    mprintf(where, " cjsw=%s ",ftos(c->cjsw,"", 7, 0));
  if (c->calc.is  &&  c->calc.rs  &&  c->calc.cj  &&  c->calc.cjsw)
    mprintf(where, "\n*+");
  if (c->calc.is  &&  c->is != NOT_INPUT)
    mprintf(where, " is=%s ", ftos(c->is,  "", 7, 0));
  if (c->calc.rs  &&  c->rs != NOT_INPUT)
    mprintf(where, " rs=%s ", ftos(c->rs,  "", 7, 0));
  if (c->calc.cj  &&  c->cj != NOT_INPUT)
    mprintf(where, " cj=%s ", ftos(c->cj,  "", 7, 0));
  if (c->calc.cjsw  &&  c->cjsw != NOT_INPUT)
    mprintf(where, " cjsw=%s ",ftos(c->cjsw,"", 7, 0));
  mprintf(where, "\n");
}
/*--------------------------------------------------------------------------*/
void DEV_DIODE::expand()
{
  DIODE_COMMON* c = (DIODE_COMMON*)common;
  assert(c);
  c->model = MODEL_BASE::rootmodel->find_model(c->modelname);
  const MODEL_DIODE* m = (const MODEL_DIODE*)c->model;
  assert(m);
  
  if (c->calc.is  ||  c->is == NOT_INPUT){
    c->is = m->js * c->area;
    c->calc.is = TRUE;
  }
  if (c->calc.rs  ||  c->rs == NOT_INPUT){
    c->rs = m->rs * c->area;
    c->calc.rs = TRUE;
  }
  if (c->calc.cj  ||  c->cj == NOT_INPUT){
    c->cj = m->cjo * c->area;
    c->calc.cj = TRUE;
  }
  if (c->calc.cjsw  ||  c->cjsw == NOT_INPUT){
    c->cjsw = m->cjsw * c->perim;
    c->calc.cjsw = TRUE;
  }

  subckt = Yj;		/* subckt points at any element in the subckt */
			/* this forces it to be one that will not be deleted */
			/* otherwise it could result in a dangling pointer */

							/* build subckt */
  if (c->cj != 0.  ||  c->cjsw != 0.){
    if (!Cj){
      Cj = new DEV_CAPACITANCE;
      subckt = Cj->insertbefore(subckt);
    }
    Cj->parent = this;
    strcpy(Cj->label,"Cj");
    Cj->n[OUT1] = n[OUT1];
    Cj->n[OUT2] = n[OUT2];
    Cj->trfun = diodecap;
    Cj->acfun = NULL;
  }else{
    delete Cj;
    Cj = NULL;
  }

  if (!Yj){
    Yj = new DEV_ADMITTANCE;
    subckt = Yj->insertbefore(subckt);
  }
  Yj->parent = this;
  strcpy(Yj->label,"Yj");
  Yj->n[OUT1] = n[OUT1];
  Yj->n[OUT2] = n[OUT2];
  Yj->trfun = evaldiode;
  Yj->acfun = NULL;
  subckt->expand_group();
}
/*--------------------------------------------------------------------------*/
double DEV_DIODE::probe_tr_num(const char *what)const
{
  CS cmd(what);
  assert(subckt);
  
  if (cmd.pmatch("Vd")){
    return n[OUT1].v0() - n[OUT2].v0();
  }else if (cmd.pmatch("Id")){
    return CARD::probe(Yj,"I") + CARD::probe(Cj,"I");
  }else if (cmd.pmatch("IJ")){
    return CARD::probe(Yj,"I");
  }else if (cmd.pmatch("IC")){
    return CARD::probe(Cj,"I");
  }else if (cmd.pmatch("P")){
    return CARD::probe(Yj,"P") + CARD::probe(Cj,"P");
  }else if (cmd.pmatch("PD")){
    return CARD::probe(Yj,"PD") + CARD::probe(Cj,"PD");
  }else if (cmd.pmatch("PS")){
    return CARD::probe(Yj,"PS") + CARD::probe(Cj,"PS");
  }else if (cmd.pmatch("PJ")){
    return CARD::probe(Yj,"P");
  }else if (cmd.pmatch("PC")){
    return CARD::probe(Cj,"P");
  }else if (cmd.pmatch("Cap")){
    return CARD::probe(Cj,"EV");
  }else if (cmd.pmatch("Req")){
    return CARD::probe(Yj,"R");
  }else if (cmd.pmatch("Y")){
    return CARD::probe(Yj,"Y") + CARD::probe(Cj,"Y");
  }else if (cmd.pmatch("Z")){
    return port_impedance(n[OUT1], n[OUT2], lu, probe_tr_num("Y"));
  }else if (cmd.pmatch("ZRAW")){
    return port_impedance(n[OUT1], n[OUT2], lu, 0.);
  }else if (cmd.pmatch("REgion")){
    return (double)region;
  }else { /* bad parameter */
    return NOT_VALID;
  }
}
/*--------------------------------------------------------------------------*/
/*static*/ void DEV_DIODE::evaldiode(CARD *brh)
{
  DEV_DIODE* p = (DEV_DIODE*)brh->parent;
  assert(p);
  const DIODE_COMMON* c = (const DIODE_COMMON*)p->common;
  assert(c);
  const MODEL_DIODE* m = (const MODEL_DIODE*)c->model;
  assert(m);

  fpoly1_t& y = brh->y0;
  double volts = y.x;
  double amps  = y.f0;
  double tempratio = SIM::temp / OPT::tnom;
  double vt = (K/Q) * SIM::temp * m->n;
  region_t oldregion = p->region;

  p->isat = c->is * pow(tempratio, m->xti) * Exp((m->eg/vt) * (tempratio-1));

  if (c->off  &&  STATUS::iter[SIM::mode] <= 1){     /* initially guess off */
    p->region = INITOFF;
    y.f1 = 0.;
    y.f0 = 0.;
    if (OPT::diodeflags & 0040){
      y.f1 = OPT::gmin;
    }
  }else if (volts < 0. /* &&  amps < 0.*/){    		  /* reverse biased */
    p->region = REVERSE;	    		  /* x = volts, f(x) = amps */
    double expterm = p->isat * Exp(volts/vt);	
    y.f0 = expterm - p->isat;        /* i = f(x) = isat * (Exp(volts/vt)-1) */
    y.f1 = expterm / vt;	     /* f'(x) = (isat/vt) * Exp(volts/vt)   */
    if (OPT::diodeflags & 0001){
      y.f1 = OPT::gmin;
    }else if (OPT::diodeflags & 0020){
      y.f1 = OPT::gmin;
      y.f0 = OPT::gmin * volts;
    }else if (OPT::diodeflags & 0100){
      y.f1 = p->isat / vt;
    }
  }else if (volts > 0.  &&  amps > 0.){			  /* forward biased */
						  /* x = amps, f(x) = volts */
    /* derivation: */		    /* if f(x) = log(u): f'(x)=(1/u)(du/dx) */
    /* poly1 r; */
    /* r.f0 = vt * log(amps/p->isat +1.); */
    /* r.f1 = vt / (isat + amps); */
    /* y.f1 = 1. / r.f1; */
    /* y.f0 = amps - r.f0*y.f1 + volts*y.f1; */
    
    p->region = FORWARD;
    y.f1 = (p->isat + amps) / vt;
    y.f0 = amps - log(amps/p->isat +1.)*(p->isat + amps) + volts*y.f1;
  }else{			    /* non-converged, inconsistent	    */
    p->region = UNKNOWN;	    /* volts and amps have different signs  */
    y.f1 = p->isat/vt;		    /* guess that the voltage should be 0   */
    y.f0 = 0.;			    /* (it usually is very close)	    */
    if (OPT::diodeflags & 0010){    /* use the correct value there	    */
      y.f0 = y.x * y.f1;
    }
  }
  if (oldregion != p->region  &&  OPT::dampstrategy & dsDEVLIMIT){
      SIM::fulldamp = TRUE;
      error(bTRACE,"%s:device limit damp\n", p->printlabel());
  }
  if (OPT::diodeflags & 0004){
    if (y.f1 < OPT::gmin)
      y.f1 = OPT::gmin;
  }
  if (OPT::diodeflags & 0002){
    y.f1 += OPT::gmin;
    y.f0 += OPT::gmin * volts;
  }
  p->gd = y.f1;
}
/*--------------------------------------------------------------------------*/
/*static*/ void DEV_DIODE::diodecap(CARD *brh)
{
  DEV_DIODE* p = (DEV_DIODE*)brh->parent;
  assert(p);
  const DIODE_COMMON* c = (const DIODE_COMMON*)p->common;
  assert(c);
  const MODEL_DIODE* m = (const MODEL_DIODE*)c->model;
  assert(m);

  double volts = brh->y0.x;
  double cd = 0;

  if (c->cj != 0.){
    if (volts < m->fcpb){
      cd = c->cj / pow(1. - (volts / m->pb),  m->mj);
    }else{
      cd = (c->cj / pow(1. - m->fc, 1. + m->mj))
	* (1. - m->fc*(1.+m->mj) + (volts/m->pb)*m->mj);
    }
  }else{
    cd = 0.;
  }
  if (c->cjsw != 0.){
    if (volts < m->fcpb){
      cd += c->cjsw / pow(1. - (volts / m->pb),  m->mjsw);
    }else{
      cd += (c->cjsw / pow(1. - m->fc, 1. + m->mjsw))
	* (1. - m->fc*(1.+m->mjsw) + (volts/m->pb)*m->mjsw);
    }
  }
  if (m->tt != 0.){
    cd += p->gd * m->tt;
  }
  brh->y0.f1 = cd;
  brh->y0.f0 = brh->y0.x * brh->y0.f1;
}
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
